from logging import Logger
from typing import Dict, List, Optional, Union

import pandas as pd

from upgini.metadata import EVAL_SET_INDEX, SORT_ID, SYSTEM_RECORD_ID, TARGET, ModelTaskType, SearchKey
from upgini.resource_bundle import ResourceBundle
from upgini.utils.datetime_utils import DateTimeSearchKeyConverter
from upgini.utils.target_utils import define_task


def remove_fintech_duplicates(
    df: pd.DataFrame,
    search_keys: Dict[str, SearchKey],
    date_format: Optional[str] = None,
    logger: Optional[Logger] = None,
    silent=False,
    bundle: ResourceBundle = None,
) -> pd.DataFrame:
    # Base checks
    date_col = _get_column_by_key(search_keys, [SearchKey.DATE, SearchKey.DATETIME])
    if define_task(df[TARGET], date_col is not None, silent=True) != ModelTaskType.BINARY:
        return df

    date_col = _get_column_by_key(search_keys, [SearchKey.DATE, SearchKey.DATETIME])
    if date_col is None:
        return df

    personal_cols = []
    phone_col = _get_column_by_key(search_keys, SearchKey.PHONE)
    if phone_col:
        personal_cols.append(phone_col)
    email_col = _get_column_by_key(search_keys, SearchKey.EMAIL)
    if email_col:
        personal_cols.append(email_col)
    hem_col = _get_column_by_key(search_keys, SearchKey.HEM)
    if hem_col:
        personal_cols.append(hem_col)
    if len(personal_cols) == 0:
        return df

    sub_df = df[personal_cols + [date_col, TARGET]]

    # Fast check for duplicates by personal keys
    if not sub_df[personal_cols].duplicated().any():
        return df

    grouped_by_personal_cols = sub_df.groupby(personal_cols, group_keys=False)

    # counts of diff dates by set of personal keys
    uniques = grouped_by_personal_cols[date_col].nunique()
    total = len(uniques)
    diff_dates = len(uniques[uniques > 1])
    if diff_dates / total >= 0.6:
        return df

    # Additional checks

    duplicates = sub_df.duplicated(personal_cols, keep=False)
    duplicate_rows = sub_df[duplicates]
    if len(duplicate_rows) == 0:
        return df

    # if there is no different target values in personal keys duplicate rows
    nonunique_target_groups = grouped_by_personal_cols[TARGET].nunique() > 1
    if nonunique_target_groups.sum() == 0:
        return df

    def has_diff_target_within_60_days(rows):
        rows = rows.sort_values(by=date_col)
        return len(rows[rows[TARGET].ne(rows[TARGET].shift()) & (rows[date_col].diff() < 60 * 24 * 60 * 60 * 1000)]) > 0

    nonunique_target_rows = nonunique_target_groups[nonunique_target_groups].reset_index().drop(columns=TARGET)
    sub_df = pd.merge(sub_df, nonunique_target_rows, on=personal_cols)

    sub_df = DateTimeSearchKeyConverter(date_col, date_format=date_format, logger=logger, bundle=bundle).convert(sub_df)
    grouped_by_personal_cols = sub_df.groupby(personal_cols, group_keys=False)
    rows_with_diff_target = grouped_by_personal_cols.filter(has_diff_target_within_60_days)
    if len(rows_with_diff_target) > 0:
        unique_keys_to_delete = rows_with_diff_target[personal_cols].drop_duplicates()
        if EVAL_SET_INDEX not in df.columns:
            rows_to_remove = pd.merge(df.reset_index(), unique_keys_to_delete, on=personal_cols)
            rows_to_remove = rows_to_remove.set_index(df.index.name or "index")
            perc = len(rows_to_remove) * 100 / len(df)
            msg = bundle.get("dataset_train_diff_target_duplicates_fintech").format(
                perc, len(rows_to_remove), rows_to_remove.index.to_list()
            )
            if not silent:
                print(msg)
            if logger:
                logger.warning(msg)
            logger.info(f"Dataset shape before clean fintech duplicates: {df.shape}")
            df = df[~df.index.isin(rows_to_remove.index)]
            logger.info(f"Dataset shape after clean fintech duplicates: {df.shape}")
        else:
            # Indices in train and eval_set can be the same so we remove rows from them separately
            train = df.query(f"{EVAL_SET_INDEX} == 0")
            train_rows_to_remove = pd.merge(train.reset_index(), unique_keys_to_delete, on=personal_cols)
            train_rows_to_remove = train_rows_to_remove.set_index(train.index.name or "index")
            train_perc = len(train_rows_to_remove) * 100 / len(train)
            msg = bundle.get("dataset_train_diff_target_duplicates_fintech").format(
                train_perc, len(train_rows_to_remove), train_rows_to_remove.index.to_list()
            )
            if not silent:
                print(msg)
            if logger:
                logger.warning(msg)
            logger.info(f"Train dataset shape before clean fintech duplicates: {train.shape}")
            train = train[~train.index.isin(train_rows_to_remove.index)]
            logger.info(f"Train dataset shape after clean fintech duplicates: {train.shape}")

            evals = [df.query(f"{EVAL_SET_INDEX} == {i}") for i in df[EVAL_SET_INDEX].unique() if i != 0]
            new_evals = []
            for i, eval in enumerate(evals):
                eval_rows_to_remove = pd.merge(eval.reset_index(), unique_keys_to_delete, on=personal_cols)
                eval_rows_to_remove = eval_rows_to_remove.set_index(eval.index.name or "index")
                eval_perc = len(eval_rows_to_remove) * 100 / len(eval)
                msg = bundle.get("dataset_eval_diff_target_duplicates_fintech").format(
                    eval_perc, len(eval_rows_to_remove), i + 1, eval_rows_to_remove.index.to_list()
                )
                if not silent:
                    print(msg)
                if logger:
                    logger.warning(msg)
                logger.info(f"Eval {i + 1} dataset shape before clean fintech duplicates: {eval.shape}")
                eval = eval[~eval.index.isin(eval_rows_to_remove.index)]
                logger.info(f"Eval {i + 1} dataset shape after clean fintech duplicates: {eval.shape}")
                new_evals.append(eval)

            logger.info(f"Dataset shape before clean fintech duplicates: {df.shape}")
            df = pd.concat([train] + new_evals)
            logger.info(f"Dataset shape after clean fintech duplicates: {df.shape}")
    return df


def clean_full_duplicates(
    df: pd.DataFrame, logger: Optional[Logger] = None, silent=False, bundle: ResourceBundle = None
) -> pd.DataFrame:
    nrows = len(df)
    if nrows == 0:
        return df
    # Remove full duplicates (exclude system_record_id, sort_id and eval_set_index)
    unique_columns = df.columns.tolist()
    if SYSTEM_RECORD_ID in unique_columns:
        unique_columns.remove(SYSTEM_RECORD_ID)
    if SORT_ID in unique_columns:
        unique_columns.remove(SORT_ID)
    if EVAL_SET_INDEX in unique_columns:
        unique_columns.remove(EVAL_SET_INDEX)
    logger.info(f"Dataset shape before clean duplicates: {df.shape}")
    # Train segment goes first so if duplicates are found in train and eval set
    # then we keep unique rows in train segment
    df = df.drop_duplicates(subset=unique_columns, keep="first")
    logger.info(f"Dataset shape after clean duplicates: {df.shape}")
    nrows_after_full_dedup = len(df)
    share_full_dedup = 100 * (1 - nrows_after_full_dedup / nrows)
    if share_full_dedup > 0:
        msg = bundle.get("dataset_full_duplicates").format(share_full_dedup)
        logger.warning(msg)
        # if not silent_mode:
        #     print(msg)
        # self.warning_counter.increment()
    if TARGET in df.columns:
        unique_columns.remove(TARGET)
        marked_duplicates = df.duplicated(subset=unique_columns, keep=False)
        if marked_duplicates.sum() > 0:
            dups_indices = df[marked_duplicates].index.to_list()
            nrows_after_tgt_dedup = len(df.drop_duplicates(subset=unique_columns, keep=False))
            num_dup_rows = nrows_after_full_dedup - nrows_after_tgt_dedup
            share_tgt_dedup = 100 * num_dup_rows / nrows_after_full_dedup

            msg = bundle.get("dataset_diff_target_duplicates").format(share_tgt_dedup, num_dup_rows, dups_indices)
            logger.warning(msg)
            if not silent:
                print(msg)
            df = df.drop_duplicates(subset=unique_columns, keep=False)
            logger.info(f"Dataset shape after clean invalid target duplicates: {df.shape}")

    return df


def _get_column_by_key(search_keys: Dict[str, SearchKey], keys: Union[SearchKey, List[SearchKey]]) -> Optional[str]:
    for col, key_type in search_keys.items():
        if (isinstance(keys, list) and key_type in keys) or key_type == keys:
            return col



from .lazy_import import LazyImport

FeaturesEnricher = LazyImport('upgini.features_enricher', 'FeaturesEnricher')
SearchKey = LazyImport('upgini.metadata', 'SearchKey')
RuntimeParameters = LazyImport('upgini.metadata', 'RuntimeParameters')
CVType = LazyImport('upgini.metadata', 'CVType')
ModelTaskType = LazyImport('upgini.metadata', 'ModelTaskType')



import logging
from typing import Optional, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from upgini.errors import ValidationError
from upgini.metadata import SYSTEM_RECORD_ID, ModelTaskType
from upgini.resource_bundle import ResourceBundle, bundle, get_custom_bundle
from upgini.sampler.random_under_sampler import RandomUnderSampler
from upgini.utils.warning_counter import WarningCounter


def correct_string_target(y: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
    if isinstance(y, pd.Series):
        return y.astype(str).astype("category").cat.codes
    elif isinstance(y, np.ndarray):
        return pd.Series(y).astype(str).astype("category").cat.codes.values


def define_task(
    y: pd.Series, has_date: bool = False, logger: Optional[logging.Logger] = None, silent: bool = False
) -> ModelTaskType:
    if logger is None:
        logger = logging.getLogger()
    target = y.dropna()
    if is_numeric_dtype(target):
        target = target.loc[np.isfinite(target)]
    else:
        target = target.loc[target != ""]
    if len(target) == 0:
        raise ValidationError(bundle.get("empty_target"))
    target_items = target.nunique()
    if target_items == 1:
        raise ValidationError(bundle.get("dataset_constant_target"))
    if target_items == 2:
        task = ModelTaskType.BINARY
    else:
        try:
            target = pd.to_numeric(target)
            is_numeric = True
        except Exception:
            is_numeric = False

        # If any value is non numeric - multiclass
        if not is_numeric:
            task = ModelTaskType.MULTICLASS
        else:
            if target.nunique() <= 50 and is_int_encoding(target.unique()):
                task = ModelTaskType.MULTICLASS
            elif has_date:
                task = ModelTaskType.REGRESSION
            else:
                non_zero_target = target[target != 0]
                target_items = non_zero_target.nunique()
                target_ratio = target_items / len(non_zero_target)
                if (
                    (target.dtype.kind == "f" and np.any(target != target.astype(int)))  # any non integer
                    or target_items > 50
                    or target_ratio > 0.2
                ):
                    task = ModelTaskType.REGRESSION
                else:
                    task = ModelTaskType.MULTICLASS

    logger.info(f"Detected task type: {task}")
    if not silent:
        print(bundle.get("target_type_detected").format(task))
    return task


def is_int_encoding(unique_values):
    return set(unique_values) == set(range(len(unique_values))) or set(unique_values) == set(
        range(1, len(unique_values) + 1)
    )


def balance_undersample(
    df: pd.DataFrame,
    target_column: str,
    task_type: ModelTaskType,
    random_state: int,
    imbalance_threshold: int = 0.2,
    min_sample_threshold: int = 5000,
    binary_bootstrap_loops: int = 5,
    multiclass_bootstrap_loops: int = 2,
    logger: Optional[logging.Logger] = None,
    bundle: Optional[ResourceBundle] = None,
    warning_counter: Optional[WarningCounter] = None,
) -> pd.DataFrame:
    if logger is None:
        logger = logging.getLogger("muted_logger")
        logger.setLevel("FATAL")
    bundle = bundle or get_custom_bundle()
    if SYSTEM_RECORD_ID not in df.columns:
        raise Exception("System record id must be presented for undersampling")

    count = len(df)
    target = df[target_column].copy()
    target_classes_count = target.nunique()

    vc = target.value_counts()
    max_class_value = vc.index[0]
    min_class_value = vc.index[len(vc) - 1]
    max_class_count = vc[max_class_value]
    min_class_count = vc[min_class_value]

    min_class_percent = imbalance_threshold / target_classes_count
    min_class_threshold = int(min_class_percent * count)

    resampled_data = df
    df = df.copy().sort_values(by=SYSTEM_RECORD_ID)
    if task_type == ModelTaskType.MULTICLASS:
        # Sort classes by rows count and find 25% quantile class
        classes = vc.index
        quantile25_idx = int(0.75 * len(classes)) - 1
        quantile25_class = classes[quantile25_idx]
        quantile25_class_cnt = vc[quantile25_class]

        if max_class_count > (quantile25_class_cnt * multiclass_bootstrap_loops):
            msg = bundle.get("imbalance_multiclass").format(quantile25_class, quantile25_class_cnt)
            logger.warning(msg)
            print(msg)
            if warning_counter:
                warning_counter.increment()

            # 25% and lower classes will stay as is. Higher classes will be downsampled
            sample_strategy = dict()
            for class_idx in range(quantile25_idx):
                # compare class count with count_of_quantile25_class * 2
                class_value = classes[class_idx]
                class_count = vc[class_value]
                sample_strategy[class_value] = min(class_count, quantile25_class_cnt * multiclass_bootstrap_loops)
            sampler = RandomUnderSampler(sampling_strategy=sample_strategy, random_state=random_state)
            X = df[SYSTEM_RECORD_ID]
            X = X.to_frame(SYSTEM_RECORD_ID)
            new_x, _ = sampler.fit_resample(X, target)  # type: ignore

            resampled_data = df[df[SYSTEM_RECORD_ID].isin(new_x[SYSTEM_RECORD_ID])]
    elif len(df) > min_sample_threshold and min_class_count < min_sample_threshold / 2:
        msg = bundle.get("dataset_rarest_class_less_threshold").format(
            min_class_value, min_class_count, min_class_threshold, min_class_percent * 100
        )
        logger.warning(msg)
        print(msg)
        if warning_counter:
            warning_counter.increment()

        # fill up to min_sample_threshold by majority class
        minority_class = df[df[target_column] == min_class_value]
        majority_class = df[df[target_column] != min_class_value]
        sample_size = min(len(majority_class), min_sample_threshold - min_class_count)
        sampled_majority_class = majority_class.sample(n=sample_size, random_state=random_state)
        resampled_data = df[
            (df[SYSTEM_RECORD_ID].isin(minority_class[SYSTEM_RECORD_ID]))
            | (df[SYSTEM_RECORD_ID].isin(sampled_majority_class[SYSTEM_RECORD_ID]))
        ]

    elif max_class_count > min_class_count * binary_bootstrap_loops:
        msg = bundle.get("dataset_rarest_class_less_threshold").format(
            min_class_value, min_class_count, min_class_threshold, min_class_percent * 100
        )
        logger.warning(msg)
        print(msg)
        if warning_counter:
            warning_counter.increment()

        sampler = RandomUnderSampler(
            sampling_strategy={max_class_value: binary_bootstrap_loops * min_class_count}, random_state=random_state
        )
        X = df[SYSTEM_RECORD_ID]
        X = X.to_frame(SYSTEM_RECORD_ID)
        new_x, _ = sampler.fit_resample(X, target)  # type: ignore

        resampled_data = df[df[SYSTEM_RECORD_ID].isin(new_x[SYSTEM_RECORD_ID])]

    logger.info(f"Shape after rebalance resampling: {resampled_data}")
    return resampled_data


def calculate_psi(expected: pd.Series, actual: pd.Series) -> float:
    df = pd.concat([expected, actual])

    # Define the bins for the target variable
    df_min = df.min()
    df_max = df.max()
    bins = [df_min, (df_min + df_max) / 2, df_max]

    # Calculate the base distribution
    train_distribution = expected.value_counts(bins=bins, normalize=True).sort_index().values

    # Calculate the target distribution
    test_distribution = actual.value_counts(bins=bins, normalize=True).sort_index().values

    # Calculate the PSI
    return np.sum((train_distribution - test_distribution) * np.log(train_distribution / test_distribution))



from typing import List

import pandas as pd

from upgini.autofe.operand import PandasOperand, VectorizableMixin


class Mean(PandasOperand, VectorizableMixin):
    name = "mean"
    output_type = "float"
    is_vector = True
    group_index = 0

    def calculate_vector(self, data: List[pd.Series]) -> pd.Series:
        return pd.DataFrame(data).T.fillna(0).mean(axis=1)


class Sum(PandasOperand, VectorizableMixin):
    name = "sum"
    is_vector = True
    group_index = 0

    def calculate_vector(self, data: List[pd.Series]) -> pd.Series:
        return pd.DataFrame(data).T.fillna(0).sum(axis=1)
